<?php

namespace Drupal\subman;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Site\Settings;
use Drupal\user\UserInterface;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Session\AccountProxy;
use Drupal\Core\Entity\EntityTypeManager;
use Symfony\Contracts\EventDispatcher\Event;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Session\AccountProxyInterface;

/**
 * Class SubmanUtilities.
 */
class SubmanUtilities implements SubmanUtilitiesInterface {

  const STORE_PREFIX = 'subman_store:';

  /**
   * The entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The config.factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected $configFactory;

  /**
   * The event.dispatcher service.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  protected $eventDispatcher;

  /**
   * The logger.factory service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The injected current_user service.
   *
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected $currentUser;

  /**
   * Constructs a new SubmanUtilities object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   * @param \Drupal\Core\Config\ConfigFactory $config_factory
   * @param \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher $event_dispatcher
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   * @param \Drupal\Core\Session\AccountProxy $current_user
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    ContainerAwareEventDispatcher $event_dispatcher,
    LoggerChannelFactoryInterface $logger_factory,
    AccountProxyInterface $current_user
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->configFactory = $config_factory;
    $this->eventDispatcher = $event_dispatcher;
    $this->loggerFactory = $logger_factory;
    $this->currentUser = $current_user;
  }

  /**
   * Returns a config value for given key.
   *
   * @param string $key
   *   The key to which to return the settings value for.
   * @param mixed $default
   *   An optional default value in case no config could be retrieved.
   *
   * @return mixed
   *   The config value retrieved.
   */
  public function config(string $key, $default = NULL) {
    $value = $default;
    $settings = Settings::get('subman.settings');
    if (is_array($settings) && isset($settings[$key])) {
      $value = $settings[$key];
    }
    return $value;
  }

  /**
   * Logs a message.
   *
   * @param string $message
   *   The message to log.
   */
  public function log(string $message, $data = NULL, array $context = [], string $level = 'info'): void {
    $sep = "<br />\n";

    // Get logger.
    $logger = $this->loggerFactory->get('subman');
    $method = method_exists($logger, $level) ? $level : 'debug';

    // Ammend caller info.
    $dbt = debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2);
    $caller = isset($dbt[1]) ? $sep . 'Caller: ' . implode(':', $dbt[1]) : '';
    $message .= $caller;

    // Add data to context and message.
    if ($data) {
      $data = is_string($data) ? $data : json_encode($data, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_LINE_TERMINATORS);
      $context['@subman_log_data'] = $data;
      $message .= $sep . 'Data:' . $sep . '<pre>@subman_log_data</pre>';
    }

    // Log message.
    $logger->{$method}($message, $context);
  }

  /**
   * Looks up an entity of given type using given key-value matches.
   *
   * @param string $entity_type
   *   The entity type to look up.
   * @param $values
   *   The [key => value, ...] match(es) to look up.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The entity found.
   */
  public function retrieveEntityByField(string $entity_type, array $values): ContentEntityInterface {
    $storage = $this->entityTypeManager->getStorage($entity_type);
    $query = $storage->getQuery();
    foreach ($values as $key => $value) {
      if (is_null($value)) {
        $query->condition($key, NULL, 'IS NULL');
      }
      else {
        $query->condition($key, $value);
      }
    }
    $query->range(0, 1);
    $ids = $query
      ->accessCheck(TRUE)
      ->execute();

    $entities = $storage->loadMultiple($ids);

    return empty($entities) ? NULL : reset($entities);
  }

  /**
   * Dispatch event reusing event_dispatcher service.
   *
   * @param string $eventName
   *   The name of the event.
   * @param \Symfony\Contracts\EventDispatcher\Event $event
   *   The event object.
   */
  public function dispatchEvent(string $eventName, Event $event = NULL): void {
    $this->eventDispatcher->dispatch($event, $eventName);
  }

  /**
   * Get request time.
   */
  public function getRequestTime(): int {
    // @todo properly inject time service.
    return \Drupal::time()->getRequestTime();
  }

  /**
   * Explodes a string into an array of separated values.
   *
   * @param string $glued_input
   *   The input string with the values still glued together.
   *
   * @return array
   *   The resulting array of separated and trimmed values.
   */
  public function explodeValues(string $glued_input): array {
    $separator = $this->config('separator', ',');
    $values = explode($separator, $glued_input);
    foreach ($values as $key => $value) {
      $values[$key] = trim($value);
    }
    return $values;
  }

  /**
   * Gets the current user.
   *
   * @param \Drupal\user\UserInterface $user
   *   Any user to return instead of current user (helps with function that take an optional user parameter).
   *
   * @return \Drupal\user\UserInterface|null
   *   Returns the current user or NULL.
   */
  public function getCurrentUser(UserInterface $user = NULL): ?UserInterface {
    if (empty($user)) {
      $user_id = $this->currentUser->id();
      $user = $this->entityTypeManager->getStorage('user')->load($user_id);
    }
    return $user;
  }

  // ---------------------------------------------------------------------------
  // Data store.
  // ---------------------------------------------------------------------------

  /**
   * Stores given value under the given key.
   *
   * @param string $key
   *   The key to store the data with.
   * @param $value
   *   The data to store.
   * @param int $expiration_time
   *   The unix timestamp after which the data is considered invalid (permanent, if nothing specified).
   */
  public function saveData(string $key, $value, int $expiration_time = Cache::PERMANENT): void {
    \Drupal::cache()->set(self::STORE_PREFIX . $key, $value, $expiration_time);
  }

  /**
   * Retrieves the stored value for the given key.
   */
  public function loadData(string $key) {
    $cache = \Drupal::cache()->get(self::STORE_PREFIX . $key);
    return $cache ? $cache->data : NULL;
  }

}
